defmodule Chronoscope.NTS.KeyEstablishment do
  import Bitwise

  @next_protocol_negotiation <<0x80, 0x01, 0x00, 0x02, 0x00, 0x00>>
  @aead_algorithm_negotiation <<0x80, 0x04, 0x00, 0x04, 0x00, 0x1E, 0x00, 0x0F>>
  @end_of_message <<0x80, 0x00, 0x00, 0x00>>

  @aead_alogorithms %{
    15 => "AEAD_AES_SIV_CMAC_256",
    30 => "AEAD_AES_128_GCM_SIV"
  }

  @next_protocols %{0 => "NTPv4"}

  def request() do
    @next_protocol_negotiation <> @aead_algorithm_negotiation <> @end_of_message
  end

  def parse_response(response) do
    {:ok, Map.update(do_parse_response(response, %{}), :cookies, 0, &length/1)}
  end

  defp do_parse_response([], acc) do
    acc
  end

  defp do_parse_response([0x80, 0x00, 0x00, 0x00], acc) do
    acc
  end

  defp do_parse_response([0x80, 0x01, length_high, length_low | rest], acc) do
    length = combine_octets(length_high, length_low)
    {next_protocols, remaining} = Enum.split(rest, length)

    do_parse_response(
      remaining,
      Map.put(acc, :next_protocols, parse_next_protocol_list(next_protocols))
    )
  end

  defp do_parse_response([first, 0x04, length_high, length_low | rest], acc)
       when first == 0x00 or first == 0x80 do
    length = combine_octets(length_high, length_low)
    {aead_algorithms, remaining} = Enum.split(rest, length)

    do_parse_response(
      remaining,
      Map.put(acc, :aead_algorithms, parse_aead_algorithm_list(aead_algorithms))
    )
  end

  defp do_parse_response([first, 0x05, length_high, length_low | rest], acc)
       when first == 0x00 or first == 0x80 do
    length = combine_octets(length_high, length_low)
    {cookie, remaining} = Enum.split(rest, length)

    do_parse_response(
      remaining,
      acc
      |> Map.update(:cookies, [cookie], &[cookie | &1])
      |> Map.put(:cookie_length, length)
    )
  end

  defp do_parse_response([_, _, length_high, length_low | rest], acc) do
    length = combine_octets(length_high, length_low)
    {_, remaining} = Enum.split(rest, length)

    do_parse_response(remaining, acc)
  end

  defp parse_aead_algorithm_list(aead_algorithms) do
    do_parse_aead_algorithm_list(aead_algorithms, [])
  end

  defp do_parse_aead_algorithm_list([], acc) do
    acc
  end

  defp do_parse_aead_algorithm_list([high, low | rest], acc) do
    @aead_alogorithms
    |> Map.get(combine_octets(high, low), "UNKNOWN")
    |> then(&do_parse_aead_algorithm_list(rest, [&1 | acc]))
  end

  defp parse_next_protocol_list(next_protocols) do
    do_parse_next_protocol_list(next_protocols, [])
  end

  defp do_parse_next_protocol_list([], acc) do
    acc
  end

  defp do_parse_next_protocol_list([high, low | rest], acc) do
    @next_protocols
    |> Map.get(combine_octets(high, low), "UNASSIGNED")
    |> then(&do_parse_aead_algorithm_list(rest, [&1 | acc]))
  end

  # todo parse server/port information

  defp combine_octets(high, low) do
    high <<< 8 ||| low
  end
end
